//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


////////////////////////////////////////////////////////////////////
// Nes Mapper #4:  MMC3
// A great majority of newer NES games (early 90's) use this mapper, both U.S and Japanese.
// Among the better-known MMC3 titles are Super Mario Bros. 2 and 3, MegaMan 3, 4, 5, and 6,
// and Crystalis
//
// Implementation Notes
// MMC3 is capable of generating IRQs baed on a scanline counting mechanism.
// The scanline counter is triggered by toggling line A13 0-->1 42 times on the PPU 
// (i.e. when switching from name table data to pattern table data during the render
// phase).  To incorporate this, we derive a subclass of the PPU VRAM that monitors
// line A13 and raises the appropriate interrupt.
// 
// In other words, a special class is derived and used in this mapper rather than piling on
// mapper-specific code into the core emulation classes.
////////////////////////////////////////////////////////////////////

#ifndef NES_MAPPER_4
#define NES_MAPPER_4

#include "../NesBaseMapper.h"
#include "../NesPPUMemory.h"
#include "../NesCPU.h"


// PRG Switching constants
#define		PRGSWITCHLOW		0x00
#define		PRGSWITCHHIGH		0x40

// CHR address XOR constants
#define		CHRNORMAL			0x00
#define		CHRXORADDR			0x80

// MMC3 command constants
#define		VROM2K0000			0x00
#define		VROM2K0800			0x01
#define		VROM1K1000			0x02
#define		VROM1K1400			0x03
#define		VROM1K1800			0x04
#define		VROM1K1C00			0x05
#define		PRGFIRST			0x06
#define		PRGLAST				0x07


class NesMapper4 : public NesBaseMapper
{
public:	
	///////////////////////////////
	// Class Procedures
	///////////////////////////////
	NesMapper4(NesCartridge *linkCartridge, NesCPU *&linkCPU, NesPPU *linkPPU);
	~NesMapper4() { };
	void initMemory();
	void write(Uint16 addr, unsigned char data);
	
	///////////////////////////////
	// Class Data Members
	///////////////////////////////
	struct {		
		unsigned char	bytPRGSwitchSlots;		// Which two 8K PRGROM slots do we swap?
		unsigned char	bytCHRAddrSel;			// XOR CHR address?
		bool			blnCHRROMFlip;
		unsigned char	bytCommand;				// Command issued by MMC3 (via $8000)		
		unsigned char	bytIRQLatch;			// IRQ Latch register
		bool			blnIRQsEnabled;			// IRQ enabled boolean
		int				intIRQCtr;				// IRQ scanline ocunter
		int				intPPUCtr;				// PPU address A13 toggle counter
	} status;

	unsigned char bytPrg0, bytPrg1, bytPrgNextLast, bytPrgLast;
	unsigned char chr01, chr23, chr4, chr5, chr6, chr7;

	void saveState(FILE *stream);
	void loadState(FILE *stream);


	char IrqReset;
	char latched;
};




// Derive a special PPU memory class that monitors the A13 line of the
// PPU read addresses, and generates an IRQ appropriately
class MMC3PPUMemory : public NesPPUMemory
{
public:
	int togctr;

	// Read: Perform some additional PPU logic, then 
	virtual Uint8 read(Uint16 &addr, bool blnCPURead) {		
		// Get a pointer to our instantiated mapper
		NesMapper4 *mmc = (NesMapper4 *) mMapper;
		
		// Increment the PPU cycle counter by 2 on every PPU VRAM access
		mmc->mPpu->timers.intCurPPUCycle += 2;

		// Perform IRQ logic
		//if (mmc->mPpu->mintCurScanLine <= 239)  {
		if (mmc->status.blnIRQsEnabled) {
			// See if line A13 toggled, and if so, decrement the A13 counter
			if ( (!(mintLastReadAddr & 0x2000)) && (addr & 0x2000) ) {
				togctr++;				

				if (mmc->status.intPPUCtr) mmc->status.intPPUCtr--;				
				if (!mmc->status.intPPUCtr) {
					togctr=0;

					// When PPU counter toggles for 42nd time, clock the IRQ counter
					if (mmc->status.intIRQCtr) mmc->status.intIRQCtr--;
										
					if (!mmc->status.intIRQCtr) {						
						// When the IRQ counter reaches zero, an IRQ is generated
						//mmc->mCpu->mintInterruptLine |= IRQ;						
						mmc->mPpu->timers.blnOtherIRQ = true;
						mmc->mPpu->timers.intOtherPix = (mmc->mPpu->timers.intCurPPUCycle);						

						//mmc->status.blnIRQsEnabled = false;
						mmc->status.intIRQCtr = mmc->status.bytIRQLatch;
					}
					
					// Reset the PPU A13 line counter
					mmc->status.intPPUCtr = 42;
					togctr = 0;
				}
			}		
		}

		// Call normal read to do core functionality
		return NesPPUMemory::read(addr, blnCPURead);
	}
};


#endif 
  