//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// NesMapper024.cpp - Implementation of mapper #24 (Konami VRC6)
//////////////////////////////////////////////////////////////////////

#include "NesMapper024.h"


NesMapper24::NesMapper24(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *&linkPPU) {
	mCartridge = linkCartridge;
	mCpu = linkCPU;
	mPpu = linkPPU;

	// Replace the standard PPU memory object with a VRC4 specific one	
	NesPPU *newPPU = new VRC6PPU();
	newPPU->Init();
	newPPU->VRAM->mbytMirroringType = linkPPU->VRAM->mbytMirroringType;
	newPPU->VRAM->link(this);	
	delete mPpu;
	mPpu = newPPU;
	linkPPU = mPpu;
	mCpu->linkCPU(mPpu, mCpu->mApu, mCpu->mController1, mCpu->mController2, this);

	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	
	initMaps();
	
	blnIrqEnable = false;
	intIrqCount = intIrqLatch = temp = 0;

	// Add a VRC6 chip to the APU
	mCpu->mApu->addExpansionChip( VRC6CHIP );
}


void NesMapper24::initMemory() {	
	if (bytPRGCount > 0) {
		swapPRGBank16(0x8000, 0);			  // First 16k bank to $8000
		swapPRGBank16(0xC000, bytPRGCount-1); // Last 16k bank to $C000
	}
}


void NesMapper24::write(Uint16 addr, unsigned char data) {	
	switch (addr) {
	case 0x8000:
		swapPRGBank16(0x8000, data % bytPRGCount);
		break;

	// ----------- VRC6 Registers ----------
	case 0x9000:
	case 0x9001:
	case 0x9002:
	case 0xA000:
	case 0xA001:
	case 0xA002:
	case 0xB000:
	case 0xB001:
	case 0xB002:
		mCpu->mApu->writeExpansionChip(addr, data);
		break;


	case 0xB003: // Mirroring
		switch (data & 0x0C) {
		case 0x00:
			mPpu->VRAM->mbytMirroringType = VERTICAL;
			break;
		case 0x04:			
			mPpu->VRAM->mbytMirroringType = HORIZONTAL;
			break;
		case 0x08:
			mPpu->VRAM->mbytMirroringType = SINGLESCREEN;
			mPpu->VRAM->mintSSMirrorLocation = 0x2000;
			break;
		case 0x0C:
			mPpu->VRAM->mbytMirroringType = SINGLESCREEN;
			mPpu->VRAM->mintSSMirrorLocation = 0x2400;
			break;
		}		
		break;

	case 0xC000:
		swapPRGBank8(0xC000, data % (bytPRGCount << 1));
		break;

	case 0xD000:
		swapCHRBank1(0x0000, data % (bytCHRCount << 3) );
		break;
	case 0xD001:
		swapCHRBank1(0x0400, data % (bytCHRCount << 3) );
		break;
	case 0xD002:
		swapCHRBank1(0x0800, data % (bytCHRCount << 3) );
		break;
	case 0xD003:
		swapCHRBank1(0x0C00, data % (bytCHRCount << 3) );
		break;
	case 0xE000:
		swapCHRBank1(0x1000, data % (bytCHRCount << 3) );
		break;
	case 0xE001:
		swapCHRBank1(0x1400, data % (bytCHRCount << 3) );
		break;
	case 0xE002:
		swapCHRBank1(0x1800, data % (bytCHRCount << 3) );
		break;
	case 0xE003:
		swapCHRBank1(0x1C00, data % (bytCHRCount << 3) );
		break;

	case 0xF000: // Set IRQ latch value
		intIrqLatch = data;
		break;
	case 0xF001: // Enable/disable IRQs
		temp = data & 0x01;
		blnIrqEnable = (data & 0x02) ? true : false;
		intIrqCount = intIrqLatch;
		break;
	case 0xF002:
		blnIrqEnable = (temp & 0x01) ? true : false;
		break;
	}
}